use lll_rs::vector::Vector;
use lll_rs::{l2::bigl2, matrix::Matrix as LLLMatrix, vector::BigVector};
use rug::Integer;
use std::cmp::min;

/// Gram-Schmidt orthogonalization
fn compute_gram_schmidt(
    basis: &LLLMatrix<BigVector>,
) -> (Vec<BigVector>, Vec<Vec<Integer>>, Vec<Integer>) {
    let n = basis.dimensions().1;
    let mut orthogonal = Vec::with_capacity(n);
    let mut mu = vec![vec![Integer::new(); n]; n];
    let mut b = Vec::with_capacity(n);

    for i in 0..n {
        let mut gs = basis[i].clone();
        for j in 0..i {
            let numerator = inner_product(&gs, &orthogonal[j]);
            let denominator: &Integer = &b[j];
            mu[i][j] = numerator.clone() / denominator.clone();

            gs = gs.sub(&orthogonal[j].clone().mulf(&mu[i][j].clone()));
        }
        let norm = inner_product(&gs, &gs);
        b.push(norm);
        orthogonal.push(gs);
    }

    (orthogonal, mu, b)
}

/// BigVectors product
fn inner_product(a: &BigVector, b: &BigVector) -> Integer {
    assert!(a.dimension() == b.dimension());
    let mut sum = Integer::new();
    for i in 0..a.dimension() {
        sum += &a[i] * &b[i];
    }
    sum
}

/// vectors orthogonal projection
fn project_block(
    basis: &LLLMatrix<BigVector>,
    start: usize,
    end: usize,
    orthogonal: &[BigVector],
    mu: &[Vec<Integer>],
) -> LLLMatrix<BigVector> {
    let mut projected = LLLMatrix::init(end - start, basis.dimensions().0);

    for (idx, i) in (start..end).enumerate() {
        let mut v = basis[i].clone();
        for (j, ortho_vec) in orthogonal.iter().enumerate().take(start) {
            let coeff = &mu[i][j];
            v = v.sub(&ortho_vec.clone().mulf(&coeff.clone()));
        }

        projected[idx] = v;
    }

    projected
}

/// BKZ reduction using LLL as SVP‐oracle
pub fn bkz_reduce(
    basis: &mut LLLMatrix<BigVector>,
    block_size: usize,
    delta: f64,
    eta: f64,
    max_rounds: usize,
) {
    let n = basis.dimensions().1;

    for _ in 0..max_rounds {
        let (orthogonal, mu, _b) = compute_gram_schmidt(basis);

        for k in 0..n {
            let l = min(k + block_size, n);
            if l - k < 2 {
                continue;
            }

            let mut projected = project_block(basis, k, l, &orthogonal, &mu);
            bigl2::lattice_reduce(&mut projected, delta, eta);

            for (i_block, i) in (k..l).enumerate() {
                basis[i] = projected[i_block].clone();
            }
        }
    }
    bigl2::lattice_reduce(basis, delta, eta);
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::matrix::Matrix;
    use rug::Integer;

    #[test]
    fn test_bkz_reduction() {
        let ciphertexts = vec![Integer::from(5), Integer::from(8), Integer::from(12)];
        let noise_bits = 2;

        let basis_matrix = Matrix::new_lattice(noise_bits, ciphertexts).unwrap();
        let mut lll_matrix = basis_matrix.to_lll_matrix();

        bkz_reduce(&mut lll_matrix, 5, 0.51, 0.75, 3);

        let shortest_vector = &lll_matrix[0];
        let mut norm = Integer::from(0);
        for i in 0..=lll_matrix.dimensions().0 - 1 {
            norm += &shortest_vector[i] * &shortest_vector[i];
        }
        assert!(
            norm > Integer::from(0),
            "Shortest vector should not be zero"
        );
    }
}
