use crate::vector::IntVector;
use rug::ops::Pow;
use rug::Integer;
use std::ops::{Index, IndexMut};

macro_rules! int {
    ($x:expr) => {
        rug::Integer::from($x)
    };
}

#[derive(Debug, PartialEq, Clone)]
pub struct Matrix {
    pub n: usize, // number of columns
    pub m: usize, // number of rows
    pub columns: Vec<IntVector>,
}

impl Matrix {
    pub fn new(n: usize, m: usize, values: Vec<Integer>) -> Option<Self> {
        if n * m != values.len() {
            return None;
        }
        let mut columns = vec![Vec::with_capacity(m); n];
        for (i, value) in values.into_iter().enumerate() {
            let col = i % n;
            columns[col].push(value);
        }
        Some(Matrix {
            n,
            m,
            columns: columns.into_iter().map(IntVector::from_vec).collect(),
        })
    }

    pub fn new_lattice(noise_bits: usize, ciphertexts: Vec<Integer>) -> Option<Self> {
        let n = ciphertexts.len();
        let mut columns = vec![Vec::with_capacity(n); n];

        columns[0].push(int!(2u64).pow((noise_bits + 1) as u32));
        for i in 1..n {
            columns[i].push(ciphertexts[i].clone());
        }

        for i in 1..n {
            for (j, column) in columns.iter_mut().enumerate().take(n) {
                if i == j {
                    column.push(ciphertexts[0].clone());
                } else {
                    column.push(int!(0));
                }
            }
        }

        Some(Matrix {
            n,
            m: n,
            columns: columns.into_iter().map(IntVector::from_vec).collect(),
        })
    }
}

impl Index<(usize, usize)> for Matrix {
    type Output = Integer;
    fn index(&self, index: (usize, usize)) -> &Self::Output {
        let (col, row) = index;
        if row >= self.n || col >= self.m {
            panic!("Matrix index out of bounds");
        }
        &self.columns[row][col]
    }
}

impl IndexMut<(usize, usize)> for Matrix {
    fn index_mut(&mut self, index: (usize, usize)) -> &mut Self::Output {
        let (col, row) = index;
        if row >= self.n || col >= self.m {
            panic!("Matrix index out of bounds");
        }
        &mut self.columns[row][col]
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::panic;

    #[test]
    fn simple_dimensions_and_index() {
        let m = Matrix::new(2, 2, vec![int!(1), int!(2), int!(3), int!(4)]).unwrap();
        assert_eq!(m.n, 2);
        assert_eq!(m.m, 2);

        // values: [1,2,3,4]
        // columns: [[1,3], [2,4]]
        assert_eq!(m[(0, 0)], int!(1));
        assert_eq!(m[(0, 1)], int!(2));
        assert_eq!(m[(1, 0)], int!(3));
        assert_eq!(m[(1, 1)], int!(4));
    }

    #[test]
    fn indexes_and_mutation() {
        let mut m = Matrix::new(2, 2, vec![int!(1), int!(2), int!(3), int!(4)]).unwrap();
        assert_eq!(m[(1, 0)], int!(3));

        m[(1, 0)] = int!(5);
        assert_eq!(m[(1, 0)], int!(5));

        let m2 = Matrix::new(
            3,
            2,
            vec![int!(1), int!(2), int!(3), int!(4), int!(5), int!(6)],
        )
        .unwrap();

        assert_eq!(m2[(0, 2)], int!(3));
        assert_eq!(m2[(1, 0)], int!(4));

        let result = panic::catch_unwind(|| {
            let _ = m2[(2, 0)];
        });
        assert!(result.is_err(), "Expected panic on m2[(2, 0)]");

        let result2 = panic::catch_unwind(|| {
            let _ = m2[(0, 3)];
        });
        assert!(result2.is_err(), "Expected panic on m2[(0, 3)]");
    }

    #[test]
    fn test_new_lattice_layout() {
        let ciphertexts = vec![int!(5), int!(8), int!(12)];
        let noise_bits = 2;
        let lattice = Matrix::new_lattice(noise_bits, ciphertexts.clone()).unwrap();

        assert_eq!(lattice.n, 3);
        assert_eq!(lattice.m, 3);

        // 1st column = [2^(noise+1), 0, 0] = [8,0,0]
        // 2nd column = [ciphertexts[1], ciphertexts[0], 0] = [8,5,0]
        // 3rd column = [ciphertexts[2], 0, ciphertexts[0]] = [12,0,5]
        let expected_flat = vec![
            int!(8),
            int!(8),
            int!(12),
            int!(0),
            int!(5),
            int!(0),
            int!(0),
            int!(0),
            int!(5),
        ];

        let mut actual_flat = Vec::with_capacity(9);
        for col in 0..lattice.m {
            for row in 0..lattice.n {
                actual_flat.push(lattice[(col, row)].clone());
            }
        }

        assert_eq!(actual_flat, expected_flat);
    }
}
